import assert from "assert"

import { DynamicBuffer } from "../ts/lib/util/DynamicBuffer"

const bytes = [0x8b, 0x62, 0x86, 0xde, 0x70, 0x0b, 0x3b, 0xfa, 0x6b, 0x78, 0xf2, 0x86, 0xd8, 0x0a, 0x0e, 0x9f, 0x9a, 0xe9, 0x96, 0x17, 0x28, 0x07, 0xc5, 0xac, 0x8e, 0x3d, 0x51, 0xa1, 0x28, 0xd8, 0x89, 0x79];
const bytestr = "8b6286de700b3bfa6b78f286d80a0e9f9ae996172807c5ac8e3d51a128d88979";

describe("DynamicBuffer", () => {
    describe("#alloc()", () => {
        it("Should allocate a dynamic buffer of a set length.", () => {
            const buf1 = DynamicBuffer.alloc(8);
            const buf2 = DynamicBuffer.alloc(-64);
            const buf3 = DynamicBuffer.alloc(2 ** 32 - 1);

            assert.strictEqual(buf1.size, 8);
            assert.strictEqual(buf2.size, 0);
            assert.strictEqual(buf3.size, 2 ** 32 - 1);

            assert.throws(() => DynamicBuffer.alloc(2 ** 32));
        });
    });
    
    describe("#from()", () => {
        it("Should create a buffer from another buffer or a string.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.toString("hex"), bytestr);
        });
    });
    
    describe("#*[Symbol.iterator]()", () => {
        it("Should loop through the bytes in a buffer.", () => {
            const buf = DynamicBuffer.from(bytes);

            let i = 0;
            for (const byte of buf) {
                assert.strictEqual(byte, bytes[i]);
                i++;
            }
        });
    });

    describe("#goto()", () => {
        it("Should go to a specified position in the buffer.", () => {
            const buf = DynamicBuffer.from(bytes);

            buf.goto(8);
            assert.strictEqual(buf.cursor, 8);
            
            buf.goto(16);
            assert.strictEqual(buf.cursor, 16);
            
            buf.goto(5);
            assert.strictEqual(buf.cursor, 5);
        });
    })

    describe("#jump()", () => {
        it("Should jump a specified number of bytes from the current position.", () => {
            const buf = DynamicBuffer.from(bytes);

            buf.jump(2);
            assert.strictEqual(buf.cursor, 2);
            
            buf.jump(4);
            assert.strictEqual(buf.cursor, 6);
            
            buf.jump(6);
            assert.strictEqual(buf.cursor, 12);
        });
    });

    describe("#realloc()", () => {
        it("Should reallocate buffer data.", () => {
            const buf = DynamicBuffer.from(bytes);

            buf.realloc(200);
            assert.strictEqual(buf.size, 200);
            
            buf.realloc(50);
            assert.strictEqual(buf.size, 50);
            
            buf.realloc(5);
            assert.strictEqual(buf.size, 5);
        });
    });

    describe("#expand()", () => {
        it("Should expand the buffer as required.", () => {
            const buf = DynamicBuffer.alloc(0);

            buf.expand(10);
            buf.expand(2);
            assert.strictEqual(buf.size, 10);
        });
    });

    describe("#uint8()", () => {
        it("Should read or write an unsigned 8-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.uint8(), 139);
            assert.strictEqual(buf.uint8(), 98);
            assert.strictEqual(buf.uint8(), 134);
            
            assert.strictEqual(buf.uint8(255), buf);
            assert.strictEqual(buf.uint8(5), buf);

            assert.throws(() => buf.uint8(256));
            assert.throws(() => buf.uint8(-50));
        });
    });

    describe("#char()", () => {
        it("Should read or write a single utf-8 character.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.char(), "");
            assert.strictEqual(buf.char(), "b");
            assert.strictEqual(buf.char(), "");
            
            assert.strictEqual(buf.char("ÿ"), buf);
            assert.strictEqual(buf.char(""), buf);

            assert.throws(() => buf.char("Ā"));
            assert.throws(() => buf.char("ￎ"));
        });
    });

    describe("#bool()", () => {
        it("Should read or write a single boolean value.", () => {
            const buf = DynamicBuffer.from([ 0x00, 0x01, 0x00 ]);

            assert.strictEqual(buf.bool(), false);
            assert.strictEqual(buf.bool(), true);
            assert.strictEqual(buf.bool(), false);
            
            assert.strictEqual(buf.bool(true), buf);
            assert.strictEqual(buf.bool(false), buf);
        });
    });

    describe("#int8()", () => {
        it("Should read or write a signed 8-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.int8(), -117);
            assert.strictEqual(buf.int8(), 98);
            assert.strictEqual(buf.int8(), -122);
            
            assert.strictEqual(buf.int8(62), buf);
            assert.strictEqual(buf.int8(-127), buf);

            assert.throws(() => buf.int8(128));
        });
    });

    describe("#uint16()", () => {
        it("Should read or write an unsigned 16-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.uint16(), 25227);
            assert.strictEqual(buf.uint16(), 56966);
            assert.strictEqual(buf.uint16("be"), 2928);
            
            assert.strictEqual(buf.uint16(65535), buf);
            assert.strictEqual(buf.uint16(12345), buf);
            
            assert.throws(() => buf.uint16(65536));
        });
    });
    
    describe("#int16()", () => {
        it("Should read or write a signed 16-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.int16(), 25227);
            assert.strictEqual(buf.int16(), -8570);
            assert.strictEqual(buf.int16("be"), 2928);
            
            assert.strictEqual(buf.int16(32767), buf);
            assert.strictEqual(buf.int16(21543), buf);
            
            assert.throws(() => buf.int16(32768));
            assert.throws(() => buf.int16(-32769));
        });
    });
    
    describe("#uint32()", () => {
        it("Should read or write an unsigned 32-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.uint32(), 3733349003);
            assert.strictEqual(buf.uint32(), 4198173552);
            assert.strictEqual(buf.uint32("be"), 2264037483);
            
            assert.strictEqual(buf.uint32(4294967295), buf);
            assert.strictEqual(buf.uint32(10000), buf);
            
            assert.throws(() => buf.uint32(4294967296));
        });
    });
    
    describe("#int32()", () => {
        it("Should read or write a signed 32-bit integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.int32(), -561618293);
            assert.strictEqual(buf.int32(), -96793744);
            assert.strictEqual(buf.int32("be"), -2030929813);
            
            assert.strictEqual(buf.int32(2147483647), buf);
            assert.strictEqual(buf.int32(1234854), buf);
            
            assert.throws(() => buf.int32(2147483648));
        });
    });
    
    describe("#float()", () => {
        it("Should read or write a floating point number.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.float(), -4841727490458124000);
            assert.strictEqual(buf.float(), -2.4279787315146363e+35);
            assert.strictEqual(buf.float("be"), -9.120722083018184e-35);
            
            assert.strictEqual(buf.float(5243785673845), buf);
            assert.strictEqual(buf.float(123), buf);
        });
    });
    
    describe("#bytes()", () => {
        it("Should read or write a specified number of bytes.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.deepStrictEqual(buf.bytes(4), [0x8b, 0x62, 0x86, 0xde]);
            assert.deepStrictEqual(buf.bytes(6), [0x70, 0x0b, 0x3b, 0xfa, 0x6b, 0x78]);
            assert.deepStrictEqual(buf.bytes(8), [0xf2, 0x86, 0xd8, 0x0a, 0x0e, 0x9f, 0x9a, 0xe9]);
            
            assert.strictEqual(buf.bytes([0x48, 0x65, 0x6c, 0x6c, 0x6f]), buf);
            assert.strictEqual(buf.bytes([0x77, 0x65, 0x61, 0x6b, 0x65, 0x79, 0x65, 0x73]), buf);
        });
    });

    describe("#string()", () => {
        it("Should read or write a string with a prefixed length.", () => {
            const buf1 = DynamicBuffer.from([0x05, 0x48, 0x65, 0x6c, 0x6c, 0x6f]);
            const buf2 = DynamicBuffer.from([0x08, 0x77, 0x65, 0x61, 0x6b, 0x65, 0x79, 0x65, 0x73]);

            assert.strictEqual(buf1.string(), "Hello");
            assert.strictEqual(buf2.string(), "weakeyes");
        });
    });

    describe("#packed()", () => {
        it("Should read or write a signed variable-sized integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.packed(), 12555);
            assert.strictEqual(buf.packed(), 1847046);
            assert.strictEqual(buf.packed(), 11);

            assert.strictEqual(buf.packed(56), buf);
            assert.strictEqual(buf.packed(543543), buf);
        });
    });

    describe("#upacked()", () => {
        it("Should read or write an unsigned variable-sized integer.", () => {
            const buf = DynamicBuffer.from(bytes);

            assert.strictEqual(buf.upacked(), 12555);
            assert.strictEqual(buf.upacked(), 1847046);
            assert.strictEqual(buf.upacked(), 11);

            assert.strictEqual(buf.upacked(-2), buf);
            assert.strictEqual(buf.upacked(4294967296), buf);
        });
    });
});